from . import initialize
initialize()

import numpy
numpy.random.seed(0)

import pickle
import tempfile
import pytest

from mhcflurry import Class1AffinityPredictor, Class1NeuralNetwork

from mhcflurry.testing_utils import cleanup, startup

# Define a fixture to initialize and clean up predictors
@pytest.fixture(scope="module")
def downloaded_predictor():
    startup()
    yield Class1AffinityPredictor.load()
    cleanup()


def predict_and_check(
        downloaded_predictor,
        allele,
        peptide,
        expected_range=(0, 500)):

    print("\n%s" % (
        downloaded_predictor.predict_to_dataframe(
            peptides=[peptide],
            allele=allele,
            include_individual_model_predictions=True)))

    (prediction,) = downloaded_predictor.predict(allele=allele, peptides=[peptide])
    assert prediction >= expected_range[0], (downloaded_predictor, prediction)
    assert prediction <= expected_range[1], (downloaded_predictor, prediction)


def test_a1_titin_epitope_downloaded_models(downloaded_predictor):
    # Test the A1 Titin epitope ESDPIVAQY from
    #   Identification of a Titin-Derived HLA-A1-Presented Peptide
    #   as a Cross-Reactive Target for Engineered MAGE A3-Directed
    #   T Cells
    predict_and_check(downloaded_predictor, "HLA-A*01:01", "ESDPIVAQY")


def test_a1_mage_epitope_downloaded_models(downloaded_predictor):
    # Test the A1 MAGE epitope EVDPIGHLY from
    #   Identification of a Titin-Derived HLA-A1-Presented Peptide
    #   as a Cross-Reactive Target for Engineered MAGE A3-Directed
    #   T Cells
    predict_and_check(downloaded_predictor, "HLA-A*01:01", "EVDPIGHLY")


def test_a2_hiv_epitope_downloaded_models(downloaded_predictor):
    # Test the A2 HIV epitope SLYNTVATL from
    #    The HIV-1 HLA-A2-SLYNTVATL Is a Help-Independent CTL Epitope
    predict_and_check(downloaded_predictor, "HLA-A*02:01", "SLYNTVATL")


def test_caching(downloaded_predictor):
    if not downloaded_predictor.allele_to_sequence:
        # Only run this test on allele-specific predictors.
        Class1NeuralNetwork.KERAS_MODELS_CACHE.clear()
        downloaded_predictor.predict(
            peptides=["SIINFEKL"],
            allele="HLA-A*02:01")
        num_cached = len(Class1NeuralNetwork.KERAS_MODELS_CACHE)
        assert num_cached > 0


def test_downloaded_predictor_is_serializable(downloaded_predictor):
    predictor_copy = pickle.loads(pickle.dumps(downloaded_predictor))
    numpy.testing.assert_equal(
        downloaded_predictor.predict(
            ["RSKERAVVVAW"], allele="HLA-A*01:01")[0],
        predictor_copy.predict(
            ["RSKERAVVVAW"], allele="HLA-A*01:01")[0])


def test_downloaded_predictor_is_savable(downloaded_predictor):
    models_dir = tempfile.mkdtemp("_models")
    print(models_dir)
    downloaded_predictor.save(models_dir)
    predictor_copy = Class1AffinityPredictor.load(models_dir)

    numpy.testing.assert_equal(
        downloaded_predictor.predict(
            ["RSKERAVVVAW"], allele="HLA-A*01:01")[0],
        predictor_copy.predict(
            ["RSKERAVVVAW"], allele="HLA-A*01:01")[0])


def test_downloaded_predictor_gives_percentile_ranks(downloaded_predictor):
    predictions = downloaded_predictor.predict_to_dataframe(
        peptides=["SAQGQFSAV", "SAQGQFSAV"],
        alleles=["HLA-A*03:01", "HLA-C*01:02"])

    print(predictions)
    assert not predictions.prediction.isnull().any()
    assert not predictions.prediction_percentile.isnull().any()





from __future__ import print_function, division, absolute_import
import collections
import logging
import sys
import os
import json

import numpy
import pandas
from mhcgnomes import parse, Allele, AlleleWithoutGene, Gene

from warnings import warn

from . import amino_acid


def normalize_allele_name(
        raw_name,
        forbidden_substrings=("MIC", "HFE"),
        raise_on_error=True,
        default_value=None):
    """
    Parses a string into a normalized allele representation.

    Parameters
    ----------
    raw_name : str
        Input string to normalize

    forbidden_substrings : tuple of str
        Fail on inputs which contain any of these strings

    raise_on_error : bool
        If an allele fails to parse raise an exception if this argument is True

    default_value : str or None
        If raise_on_error is False and allele fails to parse, return this value

    Returns
    -------
    str or None
    """
    for forbidden_substring in forbidden_substrings:
        if forbidden_substring in raw_name:
            if raise_on_error:
                raise ValueError("Unsupported gene in MHC allele name: %s" % raw_name)
            else:
                return default_value
    result = parse(
        raw_name,
        only_class1=True,
        required_result_types=[Allele, AlleleWithoutGene, Gene],
        preferred_result_types=[Allele],
        use_allele_aliases=True,
        infer_class2_pairing=False,
        collapse_singleton_haplotypes=True,
        collapse_singleton_serotypes=True,
        raise_on_error=False,
    )
    if result is None:
        if raise_on_error:
            raise ValueError("Invalid MHC allele name: %s" % raw_name)
        else:
            return default_value
    if (
        result.annotation_pseudogene
        or result.annotation_null
        or result.annotation_questionable
    ):
        if raise_on_error:
            raise ValueError("Unsupported annotation on MHC allele: %s" % raw_name)
        else:
            return default_value
    return result.restrict_allele_fields(2).to_string()


TENSORFLOW_CONFIGURED = False


def configure_tensorflow(backend=None, gpu_device_nums=None, num_threads=None):
    """
    Configure Keras backend to use GPU or CPU.

    Parameters
    ----------
    backend : string, optional
        one of 'tensorflow-default', 'tensorflow-cpu', 'tensorflow-gpu'

    gpu_device_nums : list of int, optional
        GPU devices to potentially use

    num_threads : int, optional
        Tensorflow threads to use

    """
    import tensorflow as tf
    
    # mhcflurry models use keras 2. Tensorflow now defaults to keras 3, so to load these
    # old models, we need to set the environment variable to use legacy keras. Ideally,
    # these models such be regenerated with keras 3.
    os.environ["TF_USE_LEGACY_KERAS"] = "1"
    
    global TENSORFLOW_CONFIGURED

    if TENSORFLOW_CONFIGURED:
        return
    
    if backend is not None:
        warn("Using the `backend` argument is now unused and will be deprecated; tensorflow defaults are being used.", DeprecationWarning)

    TENSORFLOW_CONFIGURED = True

    # turn on selected GPUs with memory growth enabled
    if gpu_device_nums is not None:
        physical_devices = tf.config.list_physical_devices("GPU")
        tf.config.set_visible_devices(
            [physical_devices[idx] for idx in gpu_device_nums], "GPU"
        )
        for gpu in physical_devices:
            tf.config.experimental.set_memory_growth(gpu, True)

    if num_threads:
        tf.config.threading.set_inter_op_parallelism_threads(num_threads)
        tf.config.threading.set_intra_op_parallelism_threads(num_threads)


def configure_logging(verbose=False):
    """
    Configure logging module using defaults.

    Parameters
    ----------
    verbose : boolean
        If true, output will be at level DEBUG, otherwise, INFO.
    """
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s.%(msecs)d %(levelname)s %(module)s - %(funcName)s:"
        " %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        level=level,
    )



def amino_acid_distribution(peptides, smoothing=0.0):
    """
    Compute the fraction of each amino acid across a collection of peptides.

    Parameters
    ----------
    peptides : list of string
    smoothing : float, optional
        Small number (e.g. 0.01) to add to all amino acid fractions. The higher
        the number the more uniform the distribution.

    Returns
    -------
    pandas.Series indexed by amino acids
    """
    peptides = pandas.Series(peptides)
    aa_counts = pandas.Series(peptides.map(collections.Counter).sum())
    normalized = aa_counts / aa_counts.sum()
    if smoothing:
        normalized += smoothing
        normalized /= normalized.sum()
    return normalized


def random_peptides(num, length=9, distribution=None):
    """
    Generate random peptides (kmers).

    Parameters
    ----------
    num : int
        Number of peptides to return

    length : int
        Length of each peptide

    distribution : pandas.Series
        Maps 1-letter amino acid abbreviations to
        probabilities. If not specified a uniform
        distribution is used.

    Returns
    ----------
    list of string

    """
    if num == 0:
        return []
    if distribution is None:
        distribution = pandas.Series(1, index=sorted(amino_acid.COMMON_AMINO_ACIDS))
        distribution /= distribution.sum()

    return [
        "".join(peptide_sequence)
        for peptide_sequence in numpy.random.choice(
            distribution.index, p=distribution.values, size=(int(num), int(length))
        )
    ]


def positional_frequency_matrix(peptides):
    """
    Given a set of peptides, calculate a length x amino acids frequency matrix.

    Parameters
    ----------
    peptides : list of string
        All of same length

    Returns
    -------
    pandas.DataFrame
        Index is position, columns are amino acids
    """
    length = len(peptides[0])
    assert all(len(peptide) == length for peptide in peptides)
    counts = pandas.DataFrame(
        index=[a for a in amino_acid.BLOSUM62_MATRIX.index if a != "X"],
        columns=numpy.arange(1, length + 1),
    )
    for i in range(length):
        counts[i + 1] = pandas.Series([p[i] for p in peptides]).value_counts()
    result = (counts / len(peptides)).fillna(0.0).T
    result.index.name = "position"
    return result


def save_weights(weights_list, filename):
    """
    Save model weights to the given filename using numpy's ".npz" format.

    Parameters
    ----------
    weights_list : list of numpy array

    filename : string
    """
    numpy.savez(
        filename, **dict((("array_%d" % i), w) for (i, w) in enumerate(weights_list))
    )


def load_weights(filename):
    """
    Restore model weights from the given filename, which should have been
    created with `save_weights`.

    Parameters
    ----------
    filename : string

    Returns
    ----------
    list of array
    """
    with numpy.load(filename) as loaded:
        weights = [loaded["array_%d" % i] for i in range(len(loaded.keys()))]
    return weights


class NumpyJSONEncoder(json.JSONEncoder):
    """
    JSON encoder (used with json module) that can handle numpy arrays.
    """

    def default(self, obj):
        if isinstance(
            obj,
            (
                numpy.int_,
                numpy.intc,
                numpy.intp,
                numpy.int8,
                numpy.int16,
                numpy.int32,
                numpy.int64,
                numpy.uint8,
                numpy.uint16,
                numpy.uint32,
                numpy.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(
            obj, (numpy.float_, numpy.float16, numpy.float32, numpy.float64)
        ):
            return float(obj)
        if isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



from . import initialize
initialize()

import numpy
import pandas
import pytest

from mhcflurry import Class1AffinityPredictor, Class1NeuralNetwork
from mhcflurry.common import random_peptides
from mhcflurry.downloads import get_path

from mhcflurry.testing_utils import cleanup, startup


# Define a fixture to initialize and clean up predictors
@pytest.fixture(scope="module")
def predictors():
    startup()
    predictors_dict = {
        'allele-specific': Class1AffinityPredictor.load(get_path("models_class1", "models")),
        'pan-allele': Class1AffinityPredictor.load(get_path("models_class1_pan", "models.combined"), optimization_level=0),
    }
    yield predictors_dict
    cleanup()


def test_merge(predictors):
    pan_allele_predictor = predictors['pan-allele']

    assert len(pan_allele_predictor.class1_pan_allele_models) > 1
    peptides = random_peptides(100, length=9)
    peptides.extend(random_peptides(100, length=10))
    peptides = pandas.Series(peptides).sample(frac=1.0)

    alleles = pandas.Series(
        ["HLA-A*03:01", "HLA-B*57:01", "HLA-C*02:01"]
    ).sample(n=len(peptides), replace=True)

    predictions1 = pan_allele_predictor.predict(
        peptides=peptides, alleles=alleles)

    merged = Class1NeuralNetwork.merge(
        pan_allele_predictor.class1_pan_allele_models)
    merged_predictor = Class1AffinityPredictor(
        allele_to_sequence=pan_allele_predictor.allele_to_sequence,
        class1_pan_allele_models=[merged],
    )
    predictions2 = merged_predictor.predict(peptides=peptides, alleles=alleles)
    numpy.testing.assert_allclose(predictions1, predictions2, atol=0.1)



"""
Run doctests.
"""
from . import initialize
initialize()

import os
import doctest

import pandas
import pytest


import mhcflurry
import mhcflurry.class1_presentation_predictor

os.environ["CUDA_VISIBLE_DEVICES"] = ""

from mhcflurry.testing_utils import cleanup, startup

pytest.fixture(autouse=True, scope="module")
def setup_module():
    startup()
    yield
    cleanup()


def test_doctests():
    original_precision = pandas.get_option('display.precision')
    pandas.set_option('display.precision', 3)

    results1 = doctest.testmod(mhcflurry)
    results2 = doctest.testmod(mhcflurry.class1_presentation_predictor)

    # Disabling for now until we figure out how to deal with numerical precision
    # for predictions.
    # assert results1.failed == 0, results1.failed
    # assert results2.failed == 0, results2.failed

    pandas.set_option('display.precision', original_precision)



# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import re

from setuptools import setup


readme_dir = os.path.dirname(__file__)
readme_filename = os.path.join(readme_dir, "README.md")

try:
    with open(readme_filename, "r") as f:
        readme = f.read()
except:
    logging.warning("Failed to load %s" % readme_filename)
    readme = ""


with open("mhcflurry/version.py", "r") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

if __name__ == "__main__":
    required_packages = [
        "six",
        "pandas>=0.20.3",
        "appdirs",
        "scikit-learn",
        "mhcgnomes>=0.8.4",
        "pyyaml",
        "tqdm",
        "tensorflow>=2.15.0,<2.17.0",
        "tf-keras"
    ]

    setup(
        name="mhcflurry",
        version=version,
        description="MHC Binding Predictor",
        author="Tim O'Donnell and Alex Rubinsteyn",
        author_email="timodonnell@gmail.com",
        url="https://github.com/openvax/mhcflurry",
        license="http://www.apache.org/licenses/LICENSE-2.0.html",
        entry_points={
            "console_scripts": [
                "mhcflurry-downloads = mhcflurry.downloads_command:run",
                "mhcflurry-predict = mhcflurry.predict_command:run",
                "mhcflurry-predict-scan = mhcflurry.predict_scan_command:run",
                "mhcflurry-class1-train-allele-specific-models = "
                "mhcflurry.train_allele_specific_models_command:run",
                "mhcflurry-class1-train-pan-allele-models = "
                "mhcflurry.train_pan_allele_models_command:run",
                "mhcflurry-class1-train-processing-models = "
                "mhcflurry.train_processing_models_command:run",
                "mhcflurry-class1-select-allele-specific-models = "
                "mhcflurry.select_allele_specific_models_command:run",
                "mhcflurry-class1-select-pan-allele-models = "
                "mhcflurry.select_pan_allele_models_command:run",
                "mhcflurry-class1-select-processing-models = "
                "mhcflurry.select_processing_models_command:run",
                "mhcflurry-calibrate-percentile-ranks = "
                "mhcflurry.calibrate_percentile_ranks_command:run",
                "mhcflurry-class1-train-presentation-models = "
                "mhcflurry.train_presentation_models_command:run",
                "_mhcflurry-cluster-worker-entry-point = "
                "mhcflurry.cluster_parallelism:worker_entry_point",
            ]
        },
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Operating System :: OS Independent",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ],
        package_data={
            "mhcflurry": ["downloads.yml"],
        },
        install_requires=required_packages,
        long_description=readme,
        long_description_content_type="text/markdown",
        packages=[
            "mhcflurry",
        ],
    )
